@@ -1,19 +1,57 @@
-package provide db 0.3
+package provide db 0.4.0
 
 package require Mk4tcl
 package require hook
 package require debug
 package require wa_uuid
 
-namespace eval db {
+namespace eval ::db {
+	proc __commit {{giveup 0}} {
+		if {[info exists ::db::lockingstr]} {
+			::set currlockholder [mk::get db.__webapplockingsystem!0 lockholder]
+			if {$currlockholder != $::db::lockingstr} {
+				return 0
+				debug::log db::__commit "Unable to commit changes (lost lock)"
+			}
+		}
+
+		if {[catch {
+			if {$giveup} {
+				debug::log db [list mk::set db.__webapplockingsystem!0 lockholder ""]
+				mk::set db.__webapplockingsystem!0 lockholder ""
+			}
+
+			debug::log db [list mk::file commit db]
+			mk::file commit db
+		} err]} {
+			debug::log db "   ** failed **"
+			debug::log db::__commit "Unable to commit changes ($err)"
+			return 0
+		}
+
+		return 1
+	}
+
 	# Name: ::db::disconnect
 	# Args: (none)
 	# Rets: 1 on success, 0 otherwise.
 	# Stat: In progress..
 	proc disconnect {} {
-		catch {
+		# Verify that the DB is open
+		array set opendbs [mk::file open]
+		if {[lsearch -exact [array names opendbs] db] == -1} {
+			# If not, return without making any changes.
+			return
+		}
+
+		__commit 1
+
+		if {[catch {
 			debug::log db "mk::file close db"
 			mk::file close db
+		} err]} {
+			debug::log db "  ** failed **"
+			debug::log db::disconnect "Error while closing DB: $err"
 		}
 	}
 
@@ -23,14 +61,48 @@ namespace eval db {
 	# Stat: In progress.
 	proc connect {} {
 		array set opendbs [mk::file open]
-		if {[lsearch -exact [array names opendbs] db] == -1} {
-			debug::log db "mk::file open db $::config::db(filename) -extend -nocommit"
-			mk::file open db $::config::db(filename) -extend -nocommit
+		if {[lsearch -exact [array names opendbs] db] != -1} {
+			return
+		}
+
+		if {[file exists $::config::db(filename)]} {
+			::set ::db::lockingstr [wa_uuid::gen]
+
+			# Try to open the DB with a read-write lock
+			while 1 {
+				debug::log db [list mk::file open db $::config::db(filename) -readonly]
+				mk::file open db $::config::db(filename) -readonly
+
+				::set currlockholder "FAIL"
+				catch {
+					::set currlockholder [mk::get db.__webapplockingsystem!0 lockholder]
+				}
+
+				debug::log db [list mk::file close db]
+				mk::file close db
+
+				if {$currlockholder == ""} {
+					break
+				}
+
+				debug::log db::connect "Unable to lock database, already held by $currlockholder -- waiting."
+				after [expr int(rand() * 10000)]
 
-			after idle {
-				db::disconnect
 			}
 		}
+
+		debug::log db [list mk::file open db $::config::db(filename) -nocommit]
+		mk::file open db $::config::db(filename) -nocommit
+
+		catch {
+			debug::log db [list mk::set db.__webapplockingsystem!0 lockholder $::db::lockingstr]
+			mk::set db.__webapplockingsystem!0 lockholder $::db::lockingstr
+
+			debug::log db [list mk::file commit db]
+			mk::file commit db
+		}
+
+		after idle [list db::disconnect]
 	}
 
 	# Name: ::db::create
@@ -88,9 +160,15 @@ namespace eval db {
 
 		::set dbhandle [connect]
 
+		if {[mk::view info db.__webapplockingsystem] == ""} {
+			debug::log db [list mk::view layout db.__webapplockingsystem [list lockholder:S]]
+			mk::view layout db.__webapplockingsystem [list lockholder:S]
+			mk::set db.__webapplockingsystem!0 lockholder ""
+		}
+
 		if {[mk::view info db.__unique_fields] == ""} {
-			debug::log db [list mk::view layout db.__unique_fields {database:S fields:S}]
-			mk::view layout db.__unique_fields "database:S fields:S"
+			debug::log db [list mk::view layout db.__unique_fields [list database:S fields:S]]
+			mk::view layout db.__unique_fields [list database:S fields:S]
 		}
 
 		debug::log db "mk::row append db.__unique_fields"
@@ -162,12 +240,12 @@ namespace eval db {
 			::set uniquefieldsidx [mk::select db.__unique_fields -exact database $dbname]
 
 			if {$uniquefieldsidx != ""} {
-				debug::log db      "mk::get db.__unique_fields!$uniquefieldsidx fields"
+				debug::log db "mk::get db.__unique_fields!$uniquefieldsidx fields"
 				::set uniquefields [mk::get db.__unique_fields!$uniquefieldsidx fields]
-				debug::log db "   => $uniquefields"
+				debug::log db "   ** $uniquefields **"
 			} else {
 				::set uniquefields [list]
-				debug::log db "   => (not found)"
+				debug::log db "   ** (not found) **"
 			}
 
 			foreach chkuniquefield $uniquefields {
@@ -176,7 +254,7 @@ namespace eval db {
 				}
 			}
 			if {[info exists overlappingfields]} {
-				debug::log db "  =overlap=> $overlappingfields"
+				debug::log db "   ** (overlap) $overlappingfields **"
 				::set cmdstr [list mk::select db.${dbname}]
 				foreach overlappingfield $overlappingfields {
 					lappend cmdstr -exact $overlappingfield $fieldmapping($overlappingfield)
@@ -203,8 +281,8 @@ namespace eval db {
 			debug::log db [list mk::set db.${dbname}!${idx} $fieldname $fieldvalue]
 			mk::set db.${dbname}!${idx} $fieldname $fieldvalue
 		}
-		debug::log db "mk::file commit db"
-		mk::file commit db
+
+		__commit
 
 		::set ret 1
 
@@ -263,8 +341,8 @@ namespace eval db {
 				mk::row delete db.${dbname}!${idx}
 			}
 		}
-		debug::log db "mk::file commit db"
-		mk::file commit db
+
+		__commit
 
 		::set ret 1
 
