@@ -29,11 +29,7 @@ namespace eval user {
 			return 0
 		}
 
-		if {[info exists ::cache::user([list uid $username])]} {
-			set uid $::cache::user([list uid $username])
-		} else {
-			set uid [db::get -dbname user -field uid -where user=$username]
-		}
+		set uid [db::get -dbname user -field uid -where user=$username]
 
 		if {$uid == "" || $uid == 0} {
 			return 0
@@ -194,6 +190,9 @@ namespace eval user {
 			return 0
 		}
 
+		# Invalidate cache
+		unset -nocomplain ::user::cache_get
+
 		hook::call user::delete::enter $uid
 
 		set success [db::unset -dbname user -where uid=$uid]
@@ -237,6 +236,9 @@ namespace eval user {
 			return 0
 		}
 
+		# Invalidate cache
+		unset -nocomplain ::user::cache_get
+
 		hook::call user::create::enter $uid $args
 
 		set ret 1
@@ -263,9 +265,6 @@ namespace eval user {
 			if {![hasflag "admin"] || ([lsearch -exact $flags "root"] != -1 && ![hasflag "root"])} {
 				set ret 0
 			} else {
-				# Invalidate the cache, in case we change something.
-				unset -nocomplain ::cache::user
-
 				set check [db::set -dbname user -field flags $flags -where uid=$uid]
 				if {!$check} {
 					set ret 0
@@ -320,6 +319,10 @@ namespace eval user {
 	#	UID is specified as "ALL"
 	# Stat: In progress
 	proc get args {
+		if {[info exists ::user::cache_get($args)]} {
+			return $::user::cache_get($args)
+		}
+
 		for {set idx 0} {$idx < [llength $args]} {incr idx} {
 			set curr [lindex $args $idx]
 			switch -- $curr {
@@ -382,6 +385,7 @@ namespace eval user {
 		hook::call user::get::enter $uid $fields
 
 		set ret [db::get -dbname user $wherecmd uid=$uid $fieldscmd $fields]
+		set ::user::cache_get($args) $ret
 
 		hook::call user::get::return $ret $uid $fields
 
@@ -400,11 +404,7 @@ namespace eval user {
 		}
 
 		if {[wa_uuid::type $uid] == "user"} {
-			if {[info exists ::cache::user([list flags $uid])]} {
-				set flags $::cache::user([list flags $uid])
-			} else {
-				set flags [string tolower [get -uid $uid -flags]]
-			}
+			set flags [string tolower [get -uid $uid -flags]]
 		} else {
 			set flags $uid
 		}
